import argparse
import os
import logging
import sys
from typing import Optional, Sequence

import logdriver.server

DESCRIPTION = """
Run a TCP socker server for receiving and handling pickled Python LogRecord objects
"""
MIN_PORT = 1024
MAX_PORT = 65535
PORT_DEFAULT = os.environ.get("LOGDRIVER_DEFAULT_PORT", 9079)
HOST_DEFAULT = "localhost"
LOG_LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR"]


def port_range(port: str, min_val=MIN_PORT, max_val=MAX_PORT):
    try:
        p = int(port)
        if min_val <= p <= max_val:
            return p
    except ValueError:
        pass
    raise argparse.ArgumentTypeError(
        f"Port needs to be an integer in the range {min_val}-{max_val}"
    )


def log_level(level_str: str) -> int:
    if level_str == "DEBUG":
        return logging.DEBUG
    if level_str == "INFO":
        return logging.INFO
    if level_str == "WARNING":
        return logging.WARNING
    if level_str == "ERROR":
        return logging.ERROR
    raise argparse.ArgumentTypeError(f"Invalid logging level: {level_str}")


def parse_args(args: Optional[Sequence[str]]) -> argparse.Namespace:
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument(
        "-p",
        "--port",
        type=port_range,
        metavar=f"{MIN_PORT}-{MAX_PORT}",
        help=f"the port to bind the socket server on (default {PORT_DEFAULT})",
        default=PORT_DEFAULT,
    )
    parser.add_argument(
        "-H",
        "--host",
        type=str,
        help=f"the host address the server will listen on (default {HOST_DEFAULT})",
        default=HOST_DEFAULT,
    )
    parser.add_argument(
        "-L",
        "--level",
        type=log_level,
        default=logging.WARNING,
        metavar="DEBUG, INFO, WARNING, ERROR (default WARNING)",
        help="the logging level",
    )

    return parser.parse_args(args)


def main():
    args = parse_args(sys.argv[1:])

    system_logger = logging.getLogger("logdriver.system")
    system_handler = logging.StreamHandler(stream=sys.stderr)
    system_logger.addHandler(system_handler)

    system_logger.warning("Started logdriver logging socket server")
    system_logger.warning(f"Listening for logs on {args.host}:{args.port}")
    system_logger.warning("Press CTRL+C to quit")

    user_logger = logging.getLogger("logdriver.user")
    user_handler = logging.StreamHandler(stream=sys.stdout)
    user_logger.addHandler(user_handler)

    logdriver.server.main(args.host, args.port, system_logger, user_logger)

    system_logger.warning("Bye!")



import logging
from logging import Logger
import queue
import signal
import socket
import pickle
import socketserver
import struct
from queue import Queue
from threading import Thread, Event
from time import sleep


class LogRecordStreamHandler(socketserver.StreamRequestHandler):
    def should_stop(self):
        if (
            hasattr(self.server, "stop")
            and isinstance(self.server.stop, Event)
            and self.server.stop.is_set()
        ):
            return True
        return False

    def handle(self):
        self.request.settimeout(1)
        while True:
            if self.should_stop():
                break
            try:
                chunk = self.connection.recv(4)
            except socket.timeout:
                continue
            if len(chunk) < 4:
                break
            slen = struct.unpack(">L", chunk)[0]
            chunk = self.connection.recv(slen)
            while len(chunk) < slen:
                chunk = chunk + self.connection.recv(slen - len(chunk))
            obj = pickle.loads(chunk)
            log_record = logging.makeLogRecord(obj)
            if hasattr(self.server, "queue") and isinstance(self.server.queue, Queue):
                self.server.queue.put(log_record)


class LogRecordSocketReceiver(socketserver.ThreadingTCPServer):
    def __init__(
        self,
        q: Queue,
        stop: Event,
        host: str,
        port: int,
        handler=LogRecordStreamHandler,
    ):
        socketserver.ThreadingTCPServer.__init__(self, (host, port), handler)
        self.queue = q
        self.stop = stop


class FileLogWriter(Thread):
    def __init__(self, q: Queue, stop: Event, logger: Logger):
        Thread.__init__(self)
        self.queue = q
        self.stop = stop
        self.logger = logger

    def run(self):
        while not self.stop.is_set():
            try:
                log_record = self.queue.get(timeout=1)
            except queue.Empty:
                pass
            else:
                self.logger.handle(log_record)
                self.queue.task_done()


def main(host: str, port: int, system_logger: Logger, user_logger: Logger):
    q = Queue()
    stop = Event()
    log_writer = FileLogWriter(q, stop, user_logger)
    log_writer.start()

    system_logger.warning("Starting TCP server")

    def stop_log_writer():
        stop.set()
        log_writer.join()

    try:
        tcp_server = LogRecordSocketReceiver(host=host, port=port, q=q, stop=stop)
    except OSError:
        stop_log_writer()
        raise

    def stop_server():
        stop_log_writer()
        tcp_server.shutdown()
        tcp_server.server_close()

    def shutdown(_pos, _frame):
        stop_server()

    signal.signal(signal.SIGINT, handler=shutdown)
    signal.signal(signal.SIGTERM, handler=shutdown)

    server_thread = Thread(target=tcp_server.serve_forever)
    server_thread.daemon = True
    server_thread.start()

    while not stop.is_set():
        sleep(1)

    system_logger.warning("Shutting down TCP server")
